==> nono-0.4.4.tar.gz: nono-0.4.4/host/signalthread.cpp <==
//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// シグナルスレッド
//

// このプロセス内で処理が必要な UNIX シグナルは一旦このスレッドで受け取り、
// 必要に応じてスレッドに通知などをする。
//
// スレッド作成前に StartThread() にてプロセス全体で該当シグナルをブロック
// する。他のスレッドはこの後生成するので、全スレッドがこの特性を引き継ぐ。
//
// シグナルスレッドは、該当シグナルをブロックしたままシグナルが発生するのを
// sigwait(3) で待つ (ブロックはデフォルト動作やハンドラ呼び出しが起きるのを
// ブロックするがペンディングにはなる、sigwait(3) はこのペンディングに反応
// する)。
//
// o SIGTERM を受け取ると UIMessage でメインスレッドに通知し、そこから安全に
//   プロセスを終了させる。
//
// o SIGINT(^C)、SIGTSTP(^Z)、SIGQUIT(^\) は HostCOM のうち stdio ドライバが
//   必要とするため動作を切り替える。HostCOM が stdio なら、これらのシグナル
//   は処理せず、HostCOM スレッドにパイプで通知する。
//   HostCOM が stdio 以外なら、これらのシグナルは HostCOM に通知せず、
//   こちらで処理する。
//
// o SIGHUP、SIGPIPE、SIGUSR1、SIGUSR2 はデフォルトがプロセス終了なので
//   ブロックして無視する。
//
// o SIGIO は、通常は非同期 I/O アクセスのハンドラとしてアプリケーションが
//   これを使うことが想定されており、つまりライブラリが勝手に差し替えたり
//   しないはずなので、これをこのスレッドへのメッセージ通知として使う。
//   SIGIO 発生時に特定のメンバ変数がセットされているとそれに従って
//   アクションを起こす。
//
// Linux (Ubuntu 20.04) ではシグナルの配送が POSIX とは異なっているらしく、
// プロセスに届いたシグナルを sigwait しているスレッドに配送してくれない。
// glibc のマニュアルには記載があるようだ。
// https://linuxjm.osdn.jp/html/glibc-linuxthreads/man3/pthread_kill.3.html
// Linux のマニュアルでは確認できなかった。

#include "signalthread.h"
#include "uimessage.h"
#include <climits>
#include <csignal>

// グローバル参照用
SignalThread *gSignalThread;

void signal_handler(int signo);
void
signal_handler(int signo)
{
#if defined(__linux__)
	// Linux ではプロセスに届いたシグナルを自動でスレッドに配送して
	// くれないので、自前でやる必要があるようだ。
	if (gSignalThread) {
		gSignalThread->Kill(signo);
	}
#endif
}

// コンストラクタ
SignalThread::SignalThread()
	: inherited("SignalThread")
{
	wpipe = -1;
	sigemptyset(&waitset);
}

// デストラクタ
SignalThread::~SignalThread()
{
	TerminateThread();
	gSignalThread = NULL;
}

// 常にブロックするシグナル集合
static std::vector<int> siglist_always = {
	SIGHUP,		// ignore
	SIGPIPE,	// ignore
	SIGTERM,	// プロセス終了
	SIGIO,		// コマンド通知に使う
	SIGUSR1,	// ignore
	SIGUSR2,	// ignore
};

// HostCOM が受け取るかどうかによって動作を変えるシグナル集合
static std::vector<int> siglist_com = {
	SIGINT,		// HostCOM (^C)
	SIGQUIT,	// HostCOM (^\)
	SIGTSTP,	// HostCOM (^Z)
};

// スレッド開始 (override)
//
// SIGINT(^C) など緊急性の高いシグナルのブロック期間を極力短くするため、
// シグナルのブロック(マスク)は、極力スレッドを開始する直前に行いたい。
// SignalThread は Device リストの先頭なので (see ../vm/vm.cpp)、この
// StartThread() 時点ではまだ (少なくとも自分が関与できる) 他のスレッドは
// 生成されておらず、ここでのこのスレッドに対する Mask() は、この後生成
// されるスレッドに引き継がれるベースになる。
bool
SignalThread::StartThread()
{
	for (auto signo : siglist_always) {
		Mask(SIG_BLOCK, signo);
	}
	for (auto signo : siglist_com) {
		Mask(SIG_BLOCK, signo);
	}

	// で、スレッドを開始。
	return inherited::StartThread();
}

// スレッド実行
void
SignalThread::ThreadRun()
{
	request_terminate = false;
	request_wpipe = -1;

	for (;;) {
		int signo;
		int r;

		// sigwait() は成功なら 0、失敗ならエラーコードを返す。
		putmsg(2, "sigwait...");
		r = sigwait(&waitset, &signo);
		// 終了要求が立っていたら、どのシグナルでも、エラーが起きてても
		// とにかく先にスレッドを終了する。
		if (request_terminate) {
			break;
		}
		if (r != 0) {
			if (r == EINTR) {
				continue;
			}
			putmsg(0, "sigwait: %s", strerror(r));
			// XXX どうする?
			continue;
		}
		putmsg(1, "sigwait got %d", signo);

		// ここからシグナル別の処理
		switch (signo) {
		 case SIGIO:	// コマンド
		 {
			// 本来は無視するだけなので、ついでにこのスレッドに対する
			// メッセージ通知として使用する。フラグが立ってなければ単に無視。
			//
			// SIGIO 発生時に request_wpipe が非負なら、HostCOM からの
			// パイプ接続もしくは切断の要求。
			int req = request_wpipe.exchange(-1);
			if (req >= 0) {
				if (loglevel >= 1) {
					if (req == INT_MAX) {
						putmsg(1, "SIGIO: HostCOM del");
					} else {
						putmsg(1, "SIGIO: HostCOM add %d", req);
					}
				}
				// HostCOM スレッドへのシグナル通知を変更
				ChangeCOM(req);
			}
			break;
		 }

		 case SIGTERM:	// SIGTERM なら即終了
			// メインスレッドから終了してもらう
			UIMessage::Post(UIMessage::APPEXIT);
			break;

		 case SIGINT:	// パイプがあれば配送、なければここで処理
			if (wpipe >= 0) {
				WriteSignal(wpipe, signo);
			} else {
				// メインスレッドから終了してもらう
				UIMessage::Post(UIMessage::APPEXIT);
			}
			break;

		 case SIGTSTP:	// パイプがあれば配送、なければデフォルト動作
		 case SIGQUIT:
			if (wpipe >= 0) {
				// パイプがあれば書き込んで通知。
				WriteSignal(wpipe, signo);
			} else {
				// パイプがなければ、
				// ここでデフォルト動作を起こしたい。

				// 一旦解除して..
				Mask(SIG_UNBLOCK, signo);

				// シグナルを自身に送信。
				Kill(signo);

				// そして再度ブロックしておく。
				// こうしないと、2回目の ^Z が効かないとかになる。
				Mask(SIG_BLOCK, signo);
			}
			break;

		 default:		// 他は無視する
			break;
		}
	}

	putmsg(1, "ThreadRun terminating");
}

void
SignalThread::Terminate()
{
	putmsg(1, "Terminate");

	// フラグを立ててシグナルで通知
	request_terminate = true;
	Kill(SIGIO);
}

// HostCOM スレッドへのパイプを接続する (HostCOM スレッドから呼ばれる)
void
SignalThread::AddPipe(int fd)
{
	putmsg(1, "AddPipe(fd=%d)", fd);

	if ((bool)thread == false) {
		// HostCOM から呼ばれるが、起動時の初期設定は Init() フェーズ内の
		// ため、その時点ではまだどちらもスレッドはない。
		// この場合は単にここで設定するだけでよい。
		ChangeCOM(fd);
	} else {
		// スレッド開始後なら、フラグを立ててシグナルで通知
		request_wpipe = fd;
		Kill(SIGIO);
	}
}

// HostCOM スレッドへのパイプを削除する
void
SignalThread::DeletePipe()
{
	putmsg(1, "DeletePipe(current=%d)", wpipe);

	// フラグを立ててシグナルで通知
	request_wpipe = INT_MAX;
	Kill(SIGIO);
}

// シグナルをパイプに再配送する
void
SignalThread::WriteSignal(int wfd, int signo)
{
	char buf[1];
	int r;

	buf[0] = signo;
	r = write(wpipe, buf, sizeof(buf));
	if (r < 0) {
		// エラーが起きたパイプは閉じればいいか…
		ChangeCOM(INT_MAX);
	}
}

// HostCOM スレッドとのパイプを接続/切断する。
// SignalThread 側処理。
void
SignalThread::ChangeCOM(int reqfd)
{
	// reqfd が INT_MAX なら現在のディスクリプタを切断。
	// そうでなければ、reqfd を接続。
	// そんな高頻度で投げないので運用上は問題にならないはず。

	putmsg(1, "ChangeCOM(current=%d request=%d)", wpipe, reqfd);

	if (reqfd == INT_MAX) {
		// パイプを切断してからシグナルのブロックを解除
		close(wpipe);
		wpipe = -1;

		for (auto signo : siglist_com) {
			Mask(SIG_UNBLOCK, signo);
		}
	} else {
		// 必要なシグナルをブロックしてからパイプを接続
		for (auto signo : siglist_com) {
			Mask(SIG_BLOCK, signo);
		}

		wpipe = reqfd;
	}
}

// SignalThread にシグナルを投げる (他スレッドからも使用可)
void
SignalThread::Kill(int signo)
{
	assert((bool)thread);

	putmsg(1, "Kill(%d)", signo);
	pthread_kill(thread->native_handle(), signo);
}

// SIG_BLOCK ならブロックする集合に追加 (=配送する)。
// SIG_UNBLOCK ならブロックする集合から削除 (=こちらで処理する)。
void
SignalThread::Mask(int how, int signo)
{
	const char *howstr = (how == SIG_BLOCK) ? "BLOCK" : "UNBLOCK";

	putmsg(1, "Mask(%s, %d)", howstr, signo);

	// シグナルハンドラを設定 (シグナルハンドラはプロセスグローバル)
	// signal(3) は挙動が違う可能性があるので使わないこと。
	struct sigaction act;
	memset(&act, 0, sizeof(act));
	if (how == SIG_BLOCK) {
		act.sa_handler = signal_handler;
	} else {
		act.sa_handler = SIG_DFL;
	}
	if (sigaction(signo, &act, NULL) < 0) {
		putmsg(0, "sigaction(%d, %s): %s", signo, howstr, strerror(errno));
		return;
	}

	// このスレッドのマスクを設定
	sigset_t set;
	sigemptyset(&set);
	sigaddset(&set, signo);
	if (pthread_sigmask(how, &set, NULL) < 0) {
		putmsg(0, "pthread_sigmask(%s, %d): %s",
			howstr, signo, strerror(errno));
		return;
	}

	// sigwait() 用のセットを更新
	if (how == SIG_BLOCK) {
		sigaddset(&waitset, signo);
	} else {
		sigdelset(&waitset, signo);
	}

	// 適当デバッグ表示
	if (loglevel >= 1) {
		std::string str;

#define CHECK(NAM)	do {	\
	if (sigismember(&waitset, __CONCAT(SIG,NAM))) {	\
		str += " ";		\
		str += #NAM;	\
	}	\
} while (0)
		CHECK(HUP);
		CHECK(INT);
		CHECK(QUIT);
		CHECK(PIPE);
		CHECK(TERM);
		CHECK(TSTP);
		CHECK(IO);
		CHECK(USR1);
		CHECK(USR2);

		putmsgn("waitset =%s", str.c_str());
	}
}

==> nono-0.7.2.tar.gz: nono-0.7.2/host/signalthread.cpp <==
//
// nono
// Copyright (C) 2021 nono project
// Licensed under nono-license.txt
//

//
// シグナルスレッド
//

// このプロセス内で処理が必要な UNIX シグナルは一旦このスレッドで受け取り、
// 必要に応じてスレッドに通知などをする。
//
// スレッド作成前に StartThread() にてプロセス全体で該当シグナルをブロック
// する。他のスレッドはこの後生成するので、全スレッドがこの特性を引き継ぐ。
//
// シグナルスレッドは、該当シグナルをブロックしたままシグナルが発生するのを
// sigwait(3) で待つ (ブロックはデフォルト動作やハンドラ呼び出しが起きるのを
// ブロックするがペンディングにはなる、sigwait(3) はこのペンディングに反応
// する)。
//
// o SIGTERM、SIGINT(^C) を受け取ると UIMessage でメインスレッドに通知し、
//   そこから安全にプロセスを終了させる。
//
// o SIGHUP、SIGPIPE、SIGUSR1、SIGUSR2 はデフォルトがプロセス終了なので
//   ブロックして無視する。
//
// Linux (Ubuntu 20.04) ではシグナルの配送が POSIX とは異なっているらしく、
// プロセスに届いたシグナルを sigwait しているスレッドに配送してくれない。
// glibc のマニュアルには記載があるようだ。
// https://linuxjm.osdn.jp/html/glibc-linuxthreads/man3/pthread_kill.3.html
// Linux のマニュアルでは確認できなかった。

#include "signalthread.h"
#include "uimessage.h"
#include <csignal>

static void signal_handler(int signo);

static void
signal_handler(int signo)
{
#if defined(__linux__)
	// Linux ではプロセスに届いたシグナルを自動でスレッドに配送して
	// くれないので、自前でやる必要があるようだ。
	SignalThread *sigthread = GetSignalThread();
	if (sigthread) {
		sigthread->Kill(signo);
	}
#endif
}

// コンストラクタ
SignalThread::SignalThread()
	: inherited(OBJ_SIGNAL_THREAD)
{
	sigemptyset(&waitset);
}

// デストラクタ
SignalThread::~SignalThread()
{
	TerminateThread();
}

// 常にブロックするシグナル集合
static std::vector<int> siglist_always = {
	SIGHUP,		// ignore
	SIGINT,		// プロセス終了
	SIGPIPE,	// ignore
	SIGTERM,	// プロセス終了
	SIGUSR1,	// ignore
	SIGUSR2,	// ignore
};

// スレッド開始 (override)
//
// SIGINT(^C) など緊急性の高いシグナルのブロック期間を極力短くするため、
// シグナルのブロック(マスク)は、極力スレッドを開始する直前に行いたい。
// SignalThread は Device リストの先頭なので (see ../vm/vm.cpp)、この
// StartThread() 時点ではまだ (少なくとも自分が関与できる) 他のスレッドは
// 生成されておらず、ここでのこのスレッドに対する Mask() は、この後生成
// されるスレッドに引き継がれるベースになる。
bool
SignalThread::StartThread()
{
	for (auto signo : siglist_always) {
		Mask(SIG_BLOCK, signo);
	}

	// で、スレッドを開始。
	return inherited::StartThread();
}

// スレッド実行
void
SignalThread::ThreadRun()
{
	request_terminate = false;

	for (;;) {
		int signo;
		int r;

		// sigwait() は成功なら 0、失敗ならエラーコードを返す。
		putmsg(2, "sigwait...");
		r = sigwait(&waitset, &signo);
		// 終了要求が立っていたら、どのシグナルでも、エラーが起きてても
		// とにかく先にスレッドを終了する。
		if (request_terminate) {
			break;
		}
		if (r != 0) {
			if (r == EINTR) {
				continue;
			}
			putmsg(0, "sigwait: %s", strerror(r));
			// XXX どうする?
			continue;
		}
		putmsg(1, "sigwait got %d", signo);

		// ここからシグナル別の処理
		switch (signo) {
		 case SIGTERM:
		 case SIGINT:
			// メインスレッドから終了してもらう
			UIMessage::Post(UIMessage::APPEXIT);
			break;

		 default:		// 他は無視する
			break;
		}
	}

	putmsg(1, "ThreadRun terminating");
}

void
SignalThread::Terminate()
{
	putmsg(1, "Terminate");

	// フラグを立ててシグナルで通知
	request_terminate = true;
	Kill(SIGTERM);
}

// SignalThread にシグナルを投げる (他スレッドからも使用可)
void
SignalThread::Kill(int signo)
{
	assert((bool)thread);

	putmsg(1, "Kill(%d)", signo);
	pthread_kill(thread->native_handle(), signo);
}

// SIG_BLOCK ならブロックする集合に追加 (=配送する)。
// SIG_UNBLOCK ならブロックする集合から削除 (=こちらで処理する)。
void
SignalThread::Mask(int how, int signo)
{
	const char *howstr = (how == SIG_BLOCK) ? "BLOCK" : "UNBLOCK";

	putmsg(1, "Mask(%s, %d)", howstr, signo);

	// シグナルハンドラを設定 (シグナルハンドラはプロセスグローバル)
	// signal(3) は挙動が違う可能性があるので使わないこと。
	struct sigaction act;
	memset(&act, 0, sizeof(act));
	if (how == SIG_BLOCK) {
		act.sa_handler = signal_handler;
	} else {
		act.sa_handler = SIG_DFL;
	}
	if (sigaction(signo, &act, NULL) < 0) {
		putmsg(0, "sigaction(%d, %s): %s", signo, howstr, strerror(errno));
		return;
	}

	// このスレッドのマスクを設定
	sigset_t set;
	sigemptyset(&set);
	sigaddset(&set, signo);
	if (pthread_sigmask(how, &set, NULL) < 0) {
		putmsg(0, "pthread_sigmask(%s, %d): %s",
			howstr, signo, strerror(errno));
		return;
	}

	// sigwait() 用のセットを更新
	if (how == SIG_BLOCK) {
		sigaddset(&waitset, signo);
	} else {
		sigdelset(&waitset, signo);
	}

	// 適当デバッグ表示
	if (loglevel >= 1) {
		std::string str;

#define CHECK(NAM)	do {	\
	if (sigismember(&waitset, __CONCAT(SIG,NAM))) {	\
		str += " ";		\
		str += #NAM;	\
	}	\
} while (0)
		CHECK(HUP);
		CHECK(INT);
		CHECK(PIPE);
		CHECK(TERM);
		CHECK(USR1);
		CHECK(USR2);

		putmsgn("waitset =%s", str.c_str());
	}
}
